@@ -5,18 +5,20 @@ Usage:
 
    check_links [ <pattern> ]
 
-Make sure that all of the HTTP links on the current page can be visited
+Make sure that all the HTTP links on the current page can be visited
 successfully.  If 'pattern' is given, check only URLs that match that
 regular expression.
 
 If option 'check_links.only_collect_bad_links' is on, then all bad
 links are silently collected across all calls to check_links.  The
-function 'report_bad_links' can then be used to report all of the links,
+function 'report_bad_links' can then be used to report all the links,
 together with their referring pages.
 """
 
 import re
 
+from typing import Dict, List, Set
+
 from twill import browser, commands, log, utils
 from twill.errors import TwillAssertionError
 
@@ -26,19 +28,19 @@ __all__ = ['check_links', 'report_bad_links', 'good_urls', 'bad_urls']
 if commands.options.get('check_links.only_collection_bad_links') is None:
     commands.options['check_links.only_collect_bad_links'] = False
 
-good_urls = set()
-bad_urls = dict()
+good_urls: Set[str] = set()
+bad_urls: Dict[str, Set[str]] = dict()
 
 
 def check_links(pattern=''):
     """>> check_links [<pattern>]
 
-    Make sure that all of the HTTP links on the current page can be visited
+    Make sure that all the HTTP links on the current page can be visited
     with an HTTP response 200 (success).  If 'pattern' is given, interpret
     it as a regular expression that link URLs must contain in order to be
     tested, e.g.
 
-        check_links http://.*\\.google\\.com
+        check_links https://.*\\.google\\.com
 
     would check only links to google URLs.  Note that because 'follow'
     is used to visit the pages, the referrer URL is properly set on the
@@ -56,7 +58,7 @@ def check_links(pattern=''):
     # note that in the case of duplicate URLs, only one of the
     # links is actually followed!
 
-    collected_urls = set()
+    collected_urls: Set[str] = set()
 
     links = browser.links
     if not links:
@@ -67,6 +69,7 @@ def check_links(pattern=''):
         url = link.url
         url = url.split('#', 1)[0]  # get rid of subpage pointers
 
+        # noinspection HttpUrlsUsage
         if not url.startswith(('http://', 'https://')):
             debug("url '%s' is not an HTTP link; ignoring", url)
             continue
@@ -83,7 +86,7 @@ def check_links(pattern=''):
 
     # now, for each unique and unchecked URL, follow the link
 
-    failed = []
+    failed: List[str] = []
     for url in sorted(collected_urls):
         debug("Checking %s", url)
         if url in good_urls:
@@ -107,7 +110,7 @@ def check_links(pattern=''):
 
     if commands.options['check_links.only_collect_bad_links']:
         for url in failed:
-            referrers = bad_urls.getdefault(url, set())
+            referrers = bad_urls.setdefault(url, set())
             info('*** %s', browser.url)
             referrers.add(browser.url)
     elif failed:
@@ -122,7 +125,7 @@ def check_links(pattern=''):
 def report_bad_links(fail_if_exist='true', flush_bad_links='true'):
     """>> report_bad_links [<fail-if-exist> [<flush-bad-links>]]
 
-    Report all of the links collected across check_links runs (collected
+    Report all the links collected across check_links runs (collected
     if and only if the config option check_links.only_collect_bad_links
     is set).
 
