@@ -30,10 +30,10 @@ def discard_all_messages():
 
     Set all buttons to "discard".
     """
-    _formvalue_by_regex_setall('1', '^\\d+$', '3')
+    _form_value_by_regex_setall('1', '^\\d+$', '3')
 
 
-def _formvalue_by_regex_setall(form_name, field_name, value):
+def _form_value_by_regex_setall(form_name, field_name, value):
     form = browser.form(form_name)
     if not form:
         log.error("no such form '%s'", form_name)
