@@ -6,13 +6,13 @@ added into the main twill command set.)
 
 Commands:
 
- * fv_match -- fill in *all* fields that match a regex (unlike 'formvalue'
+ * fv_match -- fill in *all* fields that match a regex (unlike 'form_value'
         which will complain about multiple matches).  Useful for forms
-        with lots of repeated fieldnames -- 'field-1', 'field-2', etc.
+        with lots of repeated field names -- 'field-1', 'field-2', etc.
 
  * fv_multi -- fill in multiple form fields at once, e.g.
 
-          fv_multi <formname> field1=value1 field2=value2 field3=value3
+          fv_multi <form_name> field1=value1 field2=value2 field3=value3
 
  * fv_multi_sub -- same as 'fv_multi', followed by a 'submit'.
 """
@@ -24,13 +24,13 @@ from twill import browser, commands, log, utils
 __all__ = ['fv_match', 'fv_multi_match', 'fv_multi', 'fv_multi_sub']
 
 
-def fv_match(form_name, regex, value):
-    """>> fv_match <formname> <field regex> <value>
+def fv_match(form_name: str, field_pattern: str, value: str) -> None:
+    """>> fv_match <form_name> <field_pattern> <value>
 
     Set value of *all* form fields with a name that matches the given
-    regular expression.
+    regular expression pattern.
 
-    (Unlike 'formvalue' or 'fv', this will not complain about multiple
+    (Unlike 'form_value' or 'fv', this will not complain about multiple
     matches!)
     """
     form = browser.form(form_name)
@@ -38,7 +38,7 @@ def fv_match(form_name, regex, value):
         log.error("no such form '%s'", form_name)
         return
 
-    regex = re.compile(regex)
+    regex = re.compile(field_pattern)
 
     matches = [ctl for ctl in form.inputs
                if regex.search(str(ctl.get('name')))]
@@ -58,19 +58,19 @@ def fv_match(form_name, regex, value):
         log.info('set %d values total', n)
 
 
-def fv_multi_match(form_name, regex, *values):
-    """>> fv_multi_match <formname> <field regex> <value> [<value> [<value>..]]
+def fv_multi_match(form_name: str, field_pattern: str, *values: str) -> None:
+    """>> fv_multi_match <form_name> <field_pattern> <value>...
 
-    Set value of each consecutive matching form field with the next specified
-    value.  If there are no more values, use the last for all remaining form
-    fields
+    Set value of each consecutive form field matching the given pattern with
+    the next specified value.  If there are no more values, use the last for
+    all remaining form fields.
     """
     form = browser.form(form_name)
     if form is None:
         log.error("no such form '%s'", form_name)
         return
 
-    regex = re.compile(regex)
+    regex = re.compile(field_pattern)
 
     matches = [
         ctl for ctl in form.inputs if regex.search(str(ctl.get('name')))]
@@ -90,29 +90,29 @@ def fv_multi_match(form_name, regex, *values):
         log.info('set %d values total', n)
 
 
-def fv_multi(form_name, *pairs):
-    """>> fv_multi <formname> [<pair1> [<pair2> [<pair3>]]]
+def fv_multi(form_name: str, *pairs: str) -> None:
+    """>> fv_multi <form_name> <pair>...
 
     Set multiple form fields; each pair should be of the form
 
-        fieldname=value
+        field_name=value
 
     The pair will be split around the first '=', and
-    'fv <formname> fieldname value' will be executed in the order the
+    'fv <form_name> field_name value' will be executed in the order the
     pairs are given.
     """
-    for p in pairs:
-        field_name, value = p.split('=', 1)
+    for pair in pairs:
+        field_name, value = pair.split('=', 1)
         commands.fv(form_name, field_name, value)
 
 
-def fv_multi_sub(form_name, *pairs):
-    """>> fv_multi_sub <formname> [<pair1> [<pair2> [<pair3>]]]
+def fv_multi_sub(form_name: str, *pairs: str) -> None:
+    """>> fv_multi_sub <form_name> <pair>...
 
     Set multiple form fields (as with 'fv_multi') and then submit().
     """
-    for p in pairs:
-        field_name, value = p.split('=', 1)
+    for pair in pairs:
+        field_name, value = pair.split('=', 1)
         commands.fv(form_name, field_name, value)
 
     commands.submit()
